import sys
import os
import socket
import nameserver

# setup logging defaults
loglevel = 0
logfile = sys.stdout

def log(level,msg):
    if level <= loglevel:
        logfile.write(msg+'\n')
        
def fixorigin(origin):
    if origin[-1:] == '.':
        return origin[:-1]
    return origin

def isipaddr(s):
    try:
        socket.inet_aton(s)
        return 1
    except socket.error:
        return 0

def hextobin(c):
    if c == '0':
        return '0000'
    elif c == '1':
        return '0001'
    elif c == '2':
        return '0010'
    elif c == '3':
        return '0011'
    elif c == '4':
        return '0100'
    elif c == '5':
        return '0101'
    elif c == '6':
        return '0110'
    elif c == '7':
        return '0111'
    elif c == '8':
        return '1000'
    elif c == '9':
        return '1001'
    elif c == 'a':
        return '1010'
    elif c == 'b':
        return '1011'
    elif c == 'c':
        return '1100'
    elif c == 'd':
        return '1101'
    elif c == 'e':
        return '1110'
    elif c == 'f':
        return '1111'

def binlify(s):
    rs = ''
    for c in s:
        rs = rs + hextobin(c)
    return rs

def innetwork(address,cidrblock):
    netaddr = binascii.hexlify(socket.inet_aton(cidrblock.split('/')[0]))
    addr = binascii.hexlify(socket.inet_aton(address))
    mask = int(cidrblock.split('/')[1])
    return binlify(addr)[:mask] == binlify(netaddr)[:mask]

class bind4parser:
    def __init__(self, configclass=nameserver.dnsconfig):
        self.configdir = '/etc/oak'
        self.cfgobj = configclass()
        self.forwarders = []
        self.xfernets = []
        self.recursion = 1
        self.lineno = 1
        self.zonedict = {}

    def ep(self):
        return 'Error on line ' + str(self.lineno) + ': '
        
    def parse(self, f):
        # parse a bind4 config file (named.boot)
        for line in f.xreadlines():
            if line[-1:] == '\n':
                line = line[:-1]
            if not line:
                continue
            tokens = line.split()
            command = tokens[0].lower().replace('/','')
            # skip comments
            if command[0] == ';':
                self.lineno = self.lineno + 1
                continue
            log(1,'COMMAND: ' + command)
            method = getattr(self, command, None)
            if method:
                method(tokens[1:])
            else:
                log(0, self.ep() + command + ' directive not implemented')
            self.lineno = self.lineno + 1
                
    def directory(self, args):
        dirname = args[0].lower()
        if os.path.isdir(dirname):
            self.configdir = args[0].lower()
        else:
            log(0,self.ep() + dirname + ' is not a directory')

    def primary(self, args):
        origin = fixorigin(args[0].lower())
        dbpath = os.path.join(self.configdir,args[1].lower())
        if not os.path.isfile(dbpath):
            log(0,self.ep() + 'zone data file ' + dbpath + 'does not exist')
            return        
        if not self.zonedict.has_key(origin):
            self.zonedict[origin] = {'origin':origin,
                                     'filename':dbpath,
                                     'type':'master'}
        else:
            log(0,self.ep() + 'duplicate zone data for ' + origin)

    def primaryin(self, args):
        self.primary(args)

    def secondary(self, args):
        self.primary(args[:-1])
        if self.zonedict.has_key(origin):
            mip = args[2].lower()
            if not isipaddr(mip):
                log(0,self.ep() + 'ip address required for ' + origin +
                    '(' + dbpath +')')
                return
            origin = fixorigin(args[0].lower())
            self.zonedict[origin]['masterip'] = mip

    def cache(self, args):
        origin = fixorigin(args[0].lower())
        dbpath = os.path.join(self.configdir,args[1].lower())
        if os.path.isfile(dbpath):
            pr = nameserver.zonefileparser()
            pr.parse(origin,dbpath)
            self.cfgobj.cached = pr.getzdict()
        else:
            log(0, self.ep() + cachefilepath + '  is not a valid path')

    def forwarders(self, args):
        for s in args:
            if isipaddr(s):
                self.forwarders.append(s)

    def options(self, args):
        # forward-only
        option = args[0].lower()
        if option == 'no-recursion':
            self.recursion = 0
        elif option == 'query-log':
            self.cfgobj.loglevel = 2
        elif option == 'forward-only':
            log(0, 'slave mode on by default')
        else:
            log(0, self.ep() + 'option ' + option + ' not supported')

    def xfernets(self, args):
        for s in args:
            i = s.find('&')
            if i == -1:
                if isipaddr(s):
                    self.xfernets.append(s)
                else:
                    log(0,self.ep() + 'invalid ip address')
            else:
                if isipaddr(s[:i]) and isipaddr(s[i+1:]):
                    self.xfernets.append(s)

    def include(self, args):
        incpath = os.path.join(self.configdir,args[0].lower())
        if os.path.isfile(incpath):
            oldlineno = self.lineno
            self.lineno = 1
            f = file(incpath)
            self.parse(f)
            f.close()
            self.lineno = oldlineno

    def inxfernets(self, ipaddr):
        foundmatch = 0
        for ipmatch in self.xfernets:
            if ipmatch.find('&') == -1:
                if ipaddr == ipmatch:
                    foundmatch = 1
            else:
                if innetwork(ipaddr, ipmatch.replace('&','/')):
                    foundmatch = 1
        return foundmatch


    def getview(self, query, address, port):
        zonelist = self.cfgobj.zonedatabase.zdict.keys()
        if query.question.qtype == 'AXFR':
            if inxfernets(address) or not self.xfernets:
                return zonelist, self.recursion, self.forwarders
            else:
                if query.question.qname == 'cache':
                    return [], 0, self.forwarders
                else:
                    return [], self.recursion, self.forwarders
        return zonelist, self.recursion, self.forwarders

    def getconfigobj(self):
        nameserver.readzonefiles(self.zonedict)
        self.cfgobj.zonedatabase = nameserver.zonedb(self.zonedict)
        self.cfgobj.getview = self.getview
        return self.cfgobj

if __name__=='__main__':
    if len(sys.argv) < 2:
        sys.exit('Usage %s: <filename>' % sys.argv[0])
    loglevel = 1
    f = file(sys.argv[1])
    psr = bind4parser()
    psr.parse(f)
    localconfig = psr.getconfigobj()
    localconfig.loglevel = 2
    nameserver.run(localconfig)
